#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Thursday, October 21st 2021, 11:26:29 pm
# Last Modified: Friday, October 22nd 2021,4:21:14 am
#####

import numpy as np
import pandas as pd
import torch

from dataclasses import dataclass
from abc import ABC, abstractmethod

from typing import List, Dict, Tuple, Union, Optional, Any

from doc_utils import DocUtils
from vectorops.constants import *
from vectorops.utils.distance_utils import *
from vectorops.dataset_manager import DatasetManager

import clip
device = "cuda" if torch.cuda.is_available() else "cpu"
model, preprocess = clip.load("ViT-B/32", device=device)

from vectorhub.base import Base2Vec
from vectorhub.encoders.text.sentence_transformers import SentenceTransformer2Vec
from vectorhub.bi_encoders.text_image.torch import ClipText2Vec

@dataclass 
class ProjectionManager(DocUtils):
    def __init__(self, 
            dm: Optional[DatasetManager] = None,
    ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        super().__init__()
        if dm: 
            self.vector_label = dm.vector_label
            self.vector_name =  dm.vector_name
            self.metadata_filename = dm.metadata_filename
            self.dataset_id = dm.dataset_id
            self.vectors = dm.vectors
            self.labels = dm.labels
            self.metadata = dm.metadata

            self.vectors_unique = dm.vectors_unique
            self.metadata_unique = dm.metadata_unique
            self.labels_unique = dm.labels_unique
            self.vectors_lut = dm.vectors_lut
            self.norm = dm.norm
        
        self.fig = None


    def _formulae_to_vector(
        self,
        formulae: List[VectorFormula], 
        encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec],
    ) -> List[np.ndarray]:
        """
        Calculate the vector of the math formulae defined by the user
        
        VectorFormula is of type
        ```
        {< VECTOR_OP >: [v1, v2, v3,  ...]}
        ```

        where VECTOR_OP is of type
        VECTOR_OPS = Literal['avg', 'sum', 'min', 'max', 'add', 'subtract', None]
        """
        formulae_vectors = []
        for formula in formulae:
            if not isinstance(formula, VectorFormula):
                raise ValueError(f'Invalid vector operation {formula} {VectorFormula}')
            op = formula.op
            queries = formula.queries
            if encoder is None:
                try:
                    if op is None:
                        f_vector = self.vectors_lut[queries[0]]
                    else:
                        qv = [self.vectors_lut[q] for q in queries]
                        f_vector = self._vector_op(operation=op, vectors=qv)
                except KeyError as e:
                    raise ValueError(f"{e} not in dataset {self.dataset_id}")

            elif (isinstance(encoder, SentenceTransformer2Vec) or isinstance(encoder, ClipText2Vec)):
                if op is None:
                    f_vector = self._encode_text(encoder, queries)
                else:
                    qv_encoded = [self._encode_text(encoder=encoder, text=q) for q in queries]
                    f_vector = self._vector_op(operation=op, vectors=qv_encoded)
            else:
                raise ValueError(f'Invalid encoder {encoder}. Must be of type [Base2Vec, SentenceTransformer2Vec, ClipText2Vec]')
                
            formulae_vectors.append(f_vector)
        return formulae_vectors


    
    def _encode_items(
        self,
        items: List[str],
        encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec] = None,
    ) -> List[np.ndarray]:  
        """
        Encodes list of items to vectors
        """
        items_encoded = []
        for i in items:
            if encoder is None:
                try:
                    items_encoded.append(self.vectors_lut[i])
                except KeyError as e:
                    raise ValueError(f"{e} not in dataset {self.dataset_id}")

            elif (isinstance(encoder, SentenceTransformer2Vec) or isinstance(encoder, ClipText2Vec)):
                text_encoded = self._encode_text(encoder, i)
                items_encoded.append(text_encoded)

        return items_encoded
        

    @staticmethod
    def _vector_op(
        operation: VectorOperation, 
        vectors: List[np.ndarray],
        axis: int = 0
    ) -> np.ndarray:
        """
        Perform vector operations (op, v1, v2)
        """
        if operation == 'avg':
            return np.mean(vectors, axis=axis)
        elif operation == 'sum':
            return np.sum(vectors, axis=axis)
        elif operation == 'min':
            return np.min(vectors, axis=axis)
        elif operation == 'max':
            return np.max(vectors, axis=axis)
        elif operation == 'add':
            return np.add.reduce(vectors, axis=axis)
        elif operation == 'subtract':
            return np.subtract.reduce(vectors, axis=axis)
        elif not isinstance(operation, VectorOperation):
            raise ValueError(f'Not a valid vector operation {operation} {VectorOperation}')


    @staticmethod
    def _encode_text(
        encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec], 
        text: str, 
        norm=True
    ) -> np.ndarray:
        '''
        Computes the text vectors for text sequence using Sentence Transformers
        '''
        text_encoded = np.array(encoder.encode(text))
        if norm: text_encoded /= np.linalg.norm(text_encoded)
        return text_encoded


    @staticmethod
    def _get_nn_idx(
        vectors: Union[pd.DataFrame, np.ndarray],
        selected_vec: Union[pd.Series, np.ndarray],
        distance_measure_mode: str = 'cosine'
    ) -> Union[pd.DataFrame, np.ndarray]:
        '''
        Gets nearest neighbours idx of a given selected vector within a vector matrix
        '''
        if (selected_vec.ndim > 1):
            selected_vec = selected_vec.iloc[0]
        
        def compare_vector(vector: np.ndarray):
            if distance_measure_mode == 'cosine':
                return cosine_dis(selected_vec, vector)
            elif distance_measure_mode == 'euclidean':
                return euc_dis(selected_vec, vector)

        if type(vectors) == pd.DataFrame:
            distance_map = vectors.apply(compare_vector, axis=1)
            return distance_map.sort_values().index
        elif type(vectors) == np.ndarray:
            distance_map = np.array([compare_vector(v) for v in vectors])
            return np.argsort(distance_map)
            # return np.argpartition(distances, range(0, k))[:k]
        else:
            raise ValueError(f'{vectors} is not valid type (pd.DataFrame or np.ndarray)')


    @abstractmethod
    def plot(self):
        pass


#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import pandas as pd

import plotly.express as px
import plotly.graph_objects as go

from dataclasses import dataclass
from doc_utils import DocUtils

from typing import List, Dict, Tuple, Union, Optional, Any

from vectorhub.base import Base2Vec

from vectorops.constants import *
from vectorops.utils.distance_utils import *
from vectorops.dataset_manager import DatasetManager
from vectorops.projection_manager import ProjectionManager

from vectorhub.base import Base2Vec
from vectorhub.encoders.text.sentence_transformers import SentenceTransformer2Vec
from vectorhub.bi_encoders.text_image.torch import ClipText2Vec

MARKER_SIZE = 5

@dataclass
class VectorDataset(DocUtils):
    dataset_id: str
    docs: Dict[str, Any]
    vector_label: str
    vector_field: str
        
    def __post_init__(self):
        self.vectors = self.get_field_across_documents(self.vector_field, self.docs)
    
    def __len__(self):
        return len(self.vectors)
    
    @property
    def vector_field_len(self):
        return len(self.vectors[0])



@dataclass
class ComparatorDataset:
    def __init__(self, 
             vectors_1: VectorDataset, 
             vectors_2: VectorDataset, 
        ):
        super().__init__()
        self.vectors_1 = vectors_1
        self.vectors_2 = vectors_2
        assert self.vectors_1.vector_field_len == self.vectors_2.vector_field_len
        self.vector_field_len = self.vectors_1.vector_field_len
    
    @property
    def vectors_1_len(self):
        return len(self.vectors_1)
    
    @property
    def vectors_2_len(self):
        return len(self.vectors_2)
    



@dataclass 
class ScatterProjection(ProjectionManager):
    def __init__(self,  
            x_axis_value: VectorOperation,
            y_axis_value: VectorOperation,
            encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec] = None,
            plot_unique: bool = False,
            plot_nn: Union[None, int] = None, 
            plot_labels: bool = False,
            dm: Optional[DatasetManager] = None,
        ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        super().__init__(dm)
        self.x_axis_value = x_axis_value
        self.y_axis_value = y_axis_value
        self.encoder = encoder
        self.plot_unique = plot_unique
        self.plot_nn = plot_nn
        self.plot_labels = plot_labels


    def _projection(
        self,
        encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec],
        vectors: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        formulae = [ self.x_axis_value, self.y_axis_value ]
        axes_vectors_list = self._formulae_to_vector(formulae=formulae, encoder=encoder)

        # if self.plot_unique:
        #     vectors_matrix = np.stack(self.vectors_unique)  # num_embeds x hidden
        # else:
        vectors_matrix = np.stack(vectors)              # num_embeds x hidden
        axes_vectors = np.stack(axes_vectors_list)      # num_axes x hidden_dimension
        
        assert axes_vectors.shape[0] == 2
        assert vectors_matrix.shape[1] == axes_vectors.shape[1]

        projection_matrix = cosine_sim_matrix(vectors_matrix, axes_vectors)
        return projection_matrix, axes_vectors


    def _build_projection_df(self,
        projection_matrix: np.ndarray,
        axes_matrix: np.ndarray,
        vectors: np.ndarray,
        labels: np.ndarray,
        vector_label: str,
        metadata_df: pd.DataFrame,
        vector_field: Union[None, str] = None,
        plot_nn: Optional[Union[None, int]] = None, 
        plot_unique: Optional[bool] = False,
        overlap_only: bool = False,
    ) -> Tuple[pd.DataFrame, List]:
        """
        Build project df from axes matrix
        """
        if overlap_only:
            '''
            Project vectors onto axes and find overlap
            '''
            search_vector_nn_idx_1 = self._get_nn_idx(vectors=vectors, selected_vec=axes_matrix[0])
            nn_1_labels = metadata_df[[vector_label]].loc[search_vector_nn_idx_1]
            
            search_vector_nn_idx_2 = self._get_nn_idx(vectors=vectors, selected_vec=axes_matrix[1])
            nn_2_labels = metadata_df[[vector_label]].loc[search_vector_nn_idx_2]

            if plot_nn:
                nn_1_labels = nn_1_labels[:int(len(labels)/2)]
                nn_1_labels['neighbours'] = 'x' if not vector_field else f'{vector_field}_x'
                nn_2_labels = nn_2_labels[:int(len(labels)/2)]
                nn_2_labels['neighbours'] = 'y' if not vector_field else f'{vector_field}_y' 
                legend = 'neighbours'
            
            labels_df =  pd.concat([ nn_1_labels, nn_2_labels ], axis=0).reset_index()      
            data_df = pd.DataFrame( projection_matrix[labels_df.index], 
                                    columns=[ str(self.x_axis_value), str(self.y_axis_value) ])

        elif plot_nn:
            '''
            Show nearest neighbours of each axis
            '''
            search_vector_nn_idx_1 = self._get_nn_idx(vectors=vectors, selected_vec=axes_matrix[0])
            nn_1_labels = metadata_df[[vector_label]].loc[search_vector_nn_idx_1][:plot_nn]
            nn_1_labels['neighbours'] = 'x' if not vector_field else f'{vector_field}_x'

            search_vector_nn_idx_2 = self._get_nn_idx(vectors=vectors, selected_vec=axes_matrix[1])
            nn_2_labels = metadata_df[[vector_label]].loc[search_vector_nn_idx_2][:plot_nn]
            nn_2_labels['neighbours'] = 'y' if not vector_field else f'{vector_field}_y' 

            labels_df =  pd.concat([nn_1_labels, nn_2_labels], axis=0).reset_index()
            data_df = pd.DataFrame(projection_matrix[labels_df.index], 
                                    columns=[ str(self.x_axis_value), str(self.y_axis_value) ])
            legend = 'neighbours'

        else:
            if plot_unique:
                labels_df = pd.DataFrame(list(set(labels)), columns=[vector_label])
            else:
                labels_df = pd.DataFrame(labels, columns=[vector_label])

            data_df = pd.DataFrame(projection_matrix, 
                                    columns=[ str(self.x_axis_value), str(self.y_axis_value) ])
            legend = None

        return pd.concat([labels_df, data_df], axis=1), legend



    def get_neighbours_info(self,
        columns: Optional[Union[List, None]] = None,
        axis: Literal['x', 'y'] = 'x',
    ):
        if self.df is None:
            raise ValueError(f'Please call `pm.plot_scatter()`')
        if columns:
            columns = [self.vector_label] + columns
        else:
            columns = self.metadata.columns
        return self.metadata[columns].loc[self.df[self.df.neighbours==axis]['index']]


    def plot(
        self,
        plot_unique: Optional[bool] = False,
        plot_nn: Optional[Union[None, int]] = None, 
        plot_labels: Optional[bool] = False,
        encoder: Optional[Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec]] = None,
    ):
        """
        Plots scatter plot of the projection of the dataset
        """
        self.plot_unique = plot_unique
        self.plot_nn = plot_nn
        self.plot_labels = plot_labels
        self.encoder = encoder
        
        """
        Plotting projection matrix
        """
        self.projection_matrix, self.axes_matrix = self._projection(vectors=self.vectors, encoder=encoder)
        
        """
        Preparing plot figure
        """
        df, legend = self._build_projection_df(
            projection_matrix=self.projection_matrix, 
            axes_matrix=self.axes_matrix,
            vectors=self.vectors,
            labels=self.labels,
            vector_label=self.vector_label,
            metadata_df=self.metadata,
            plot_nn=self.plot_nn,
            plot_unique=self.plot_unique
        )
        text = df[self.vector_label] if (self.plot_labels) else None

        plot_title = f"{self.dataset_id}: {len(df)} points<br>{self.vector_label}: {self.vector_name}"
        plot_title += f"<br>Encoder: {self.encoder.__name__}"
        if self.plot_unique: 
            plot_title += f"\tUnique: {self.plot_unique}"

        fig = px.scatter(df, 
                        x=str(self.x_axis_value), y=str(self.y_axis_value), 
                        color=legend, text=text,
                        title=plot_title
                        )
        fig.update_traces(
            hovertemplate="<br>".join([
                "X: %{x}",
                "Y: %{y}",
                "Label: %{customdata}",
            ])
        )
    
        """
        Adding vector label to hover data
        """
        fig.update_traces(customdata=df[self.vector_label])
        fig.update_traces(hovertemplate='%{customdata}')
        fig.update_traces(
            hovertemplate="<br>".join([
                "X: %{x}",
                "Y: %{y}",
                f"{self.vector_label}: %{{customdata}}",
            ])
        )
        fig.update_layout(
            title={
                "text": plot_title,
                "y": 0.9,
                "x": 0.05,
                "xanchor": "left",
                "yanchor": "top",
                "font": {"size": 12},
            },
            xaxis_title=str(self.x_axis_value),
            yaxis_title=str(self.y_axis_value),
        )

        """
        Show trendline
        """
        projection_matrix = np.array(df[[str(self.x_axis_value), str(self.y_axis_value)]])
        fig.add_trace(
            go.Scatter(
                x=[projection_matrix.min(), projection_matrix.max()],
                y=[projection_matrix.min(), projection_matrix.max()],
                mode="lines",
                line=go.scatter.Line(color="gray"),
                showlegend=False
            )
        )
        self.fig = fig
        self.df = df
        
        return self.fig

    
    def _generate_hover_template(
        self, 
        df: pd.DataFrame,
        labels: List[str]
    ) -> Tuple[Optional[pd.DataFrame], Optional[str]]:
        """
        Generating hover template
        """
        custom_data = df[labels]
        custom_data_hover = [
            f"{c}: %{{customdata[{i}]}}"
            for i, c in enumerate(labels)
        ]
        hovertemplate = (
            "<br>".join(
                ["X: %{x}   Y: %{y}"]
                + custom_data_hover
            )
            + "<extra></extra>"
        )

        return custom_data, hovertemplate

        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import pandas as pd

import plotly.express as px
import plotly.graph_objects as go

from dataclasses import dataclass
from doc_utils import DocUtils

from typing import List, Dict, Tuple, Union, Optional, Any

from vectorops.constants import *
from vectorops.utils.distance_utils import *
from vectorops.dataset_manager import DatasetManager
from vectorops.projection_manager import ProjectionManager
from vectorops.projection.scatter import ScatterProjection

from vectorhub.base import Base2Vec
from vectorhub.encoders.text.sentence_transformers import SentenceTransformer2Vec
from vectorhub.bi_encoders.text_image.torch import ClipText2Vec

MARKER_SIZE = 8

@dataclass
class VectorDataset(DocUtils):
    dataset_id: str
    docs: Dict[str, Any]
    vector_label: str
    vector_field: str
    encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec]
    norm: bool
        
    def __post_init__(self):
        self.vectors = np.array(self.get_field_across_documents(self.vector_field, self.docs, missing_treatment='raise_error'))
        if self.norm: self.vectors /= np.linalg.norm(self.vectors, axis=1).reshape(-1, 1)
        self.labels = np.array(self.get_field_across_documents(self.vector_label, self.docs, missing_treatment='raise_error'))
    
    def __len__(self):
        return len(self.vectors)
    
    @property
    def vector_field_len(self):
        return len(self.vectors[0])



# @dataclass
# class ComparatorDataset(DocUtils):
#     def __init__(self, 
#              vectors_1: VectorDataset, 
#              vectors_2: VectorDataset, 
#         ):
#         super().__init__()
#         self.vectors_1 = vectors_1
#         self.vectors_2 = vectors_2
#         assert self.vectors_1.vector_field_len == self.vectors_2.vector_field_len
#         self.vector_field_len = self.vectors_1.vector_field_len
    
#     @property
#     def vectors_1_len(self):
#         return len(self.vectors_1)
    
#     @property
#     def vectors_2_len(self):
#         return len(self.vectors_2)
    



@dataclass 
class EmbeddingComparator(ScatterProjection):
    def __init__(self,  
            x_axis_value: VectorOperation,
            y_axis_value: VectorOperation,
            encoder: Union[Base2Vec, SentenceTransformer2Vec, ClipText2Vec] = None,
            plot_unique: bool = False,
            plot_nn: Union[None, int] = None, 
            plot_labels: bool = False,
            dm: Optional[DatasetManager] = None,
        ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        self.x_axis_value = x_axis_value
        self.y_axis_value = y_axis_value
        self.encoder = encoder
        self.plot_unique = plot_unique
        self.plot_nn = plot_nn
        self.plot_labels = plot_labels
        super().__init__(dm=dm, 
                x_axis_value=x_axis_value, y_axis_value=y_axis_value, 
                encoder=encoder, plot_unique=plot_unique, plot_nn=plot_nn, plot_labels=plot_labels
                )

    
    def compare_vector_fields(
        self,
        dataset_id: str,
        docs: List[Dict[str, Any]],
        vector_field_1: str,
        vector_field_2: str,
        vector_label: str,
        encoder_1: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = SentenceTransformer2Vec('msmarco-distilbert-dot-v5'),
        encoder_2: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = SentenceTransformer2Vec('msmarco-distilbert-dot-v5'),
        norm: bool = True,
        norm_range: bool = True,

        vector_label_char_length: Union[None, int] = 20,
        plot_nn: Optional[Union[None, int]] = None, 
        plot_labels: Optional[bool] = False,
        overlap_only: bool = True,
        
        sequences: Optional[List[str]] = None,
        marker_size: int = MARKER_SIZE,
    ):
        """
        Plots scatter plot of the projection of the dataset
        """
        self.dataset_id = dataset_id
        self.vector_label_char_length = vector_label_char_length
        self.plot_nn = plot_nn
        self.plot_labels = plot_labels

        palette = ['rgb(59,216,195)', 'rgb(251,195,51)']
        plot_mode = "markers+text" if plot_labels else "markers"
        plot_title = f"<b>Embedding Comparator Plot<br></b>"
        plot_title = plot_title.replace("</b>", f"Dataset: {dataset_id}<br></b>")
        plot_title = plot_title.replace("</b>", f"Vector Label: {vector_label}<br></b>")
        if plot_nn:
            if overlap_only: 
                plot_title = plot_title.replace("</b>", f"NN: {plot_nn}</b>\t")
            else:
                plot_title = plot_title.replace("</b>", f"NN from each axis: {plot_nn}</b>\t")
        
        vectors_1 = VectorDataset(dataset_id=dataset_id, docs=docs, 
                                vector_label=vector_label, vector_field=vector_field_1, encoder=encoder_1, norm=norm)
        vectors_2 = VectorDataset(dataset_id=dataset_id, docs=docs, 
                                vector_label=vector_label, vector_field=vector_field_2, encoder=encoder_2, norm=norm)
        labels = vectors_1.labels
        label_df = pd.DataFrame(docs, columns=[vector_label])

        data = []
        comparator_df = pd.DataFrame()
        """
        Compute axes matrix
        """
        for vector_dataset in [ vectors_1, vectors_2 ]:
            """
            Plotting projection matrix
            """
            vectors = vector_dataset.vectors
            if norm_range:
                vectors =  min_max_norm( vectors, vectors.min(), vectors.max() )

            encoder = vector_dataset.encoder
            vector_field = vector_dataset.vector_field
            vector_field_len = vector_dataset.vector_field_len

            projection_matrix, axes_matrix = self._projection( vectors=vectors, encoder=encoder )
            if norm_range:
                projection_matrix = min_max_norm( projection_matrix, projection_matrix.min(), projection_matrix.max() )
                axes_matrix = min_max_norm( axes_matrix, axes_matrix.min(), axes_matrix.max() )

            """
            Preparing plot figure
            """
            df, legend = self._build_projection_df(
                projection_matrix=projection_matrix, 
                axes_matrix=axes_matrix,
                vectors=vectors,
                labels=labels,
                vector_label=vector_label,
                vector_field=vector_field,
                metadata_df=label_df,
                plot_nn=plot_nn,
                # plot_unique=self.plot_unique,
                overlap_only=overlap_only
            )
            
            if plot_nn:
                groups = df.groupby(legend)
                for idx, val in groups:
                    text = val[vector_label] if (plot_labels) else None
                    if self.vector_label_char_length:
                        text = val[vector_label].apply( lambda x: x[: self.vector_label_char_length ] + "..." )
                    
                    custom_data, hovertemplate = self._generate_hover_template(df=val, labels=[vector_label, legend])
                    scatter = go.Scatter(
                            name=idx,
                            x=val[str(self.x_axis_value)],
                            y=val[str(self.y_axis_value)],
                            mode=plot_mode,
                            marker={"size": marker_size, "symbol": "circle"},
                            text=text,
                            textposition="middle center",
                            customdata=custom_data,
                            hovertemplate=hovertemplate,
                        )
                    data.append(scatter)
                    
            else:
                # df = df.rename(columns={vector_label: vector_field})
                text = df[vector_label] if (plot_labels) else None
                if self.vector_label_char_length:
                    text = df[vector_label].apply( lambda x: x[: self.vector_label_char_length ] + "..." )
                
                custom_data, hovertemplate = self._generate_hover_template(df=df, labels=[vector_label])
                scatter = go.Scatter(
                        name=vector_field,
                        x=df[str(self.x_axis_value)],
                        y=df[str(self.y_axis_value)],
                        mode=plot_mode,
                        marker={"size": marker_size, "symbol": "circle"},
                        text=text,
                        textposition="middle center",
                        customdata=custom_data,
                        hovertemplate=hovertemplate,
                    )
                data.append(scatter)
            
            if (not comparator_df.empty) and overlap_only:
                data = []
                overlap_df = pd.merge(comparator_df, df, on=[vector_label], how="inner")
                
                vectors_1_df = overlap_df[[vector_label, f'{str(self.x_axis_value)}_x', f'{str(self.y_axis_value)}_x']]
                vectors_1_df = vectors_1_df.rename(columns={
                    f'{str(self.x_axis_value)}_x': str(self.x_axis_value), 
                    f'{str(self.y_axis_value)}_x': str(self.y_axis_value), 
                    }
                )
                vectors_1_df = vectors_1_df.groupby([vector_label], sort=False).mean().reset_index()
                vectors_1_df['vector_field'] = vector_field_1
                if plot_nn: 
                    vectors_1_df['neighbours'] = overlap_df[['neighbours_x']]
                    vectors_1_df = vectors_1_df[:plot_nn]

                vectors_2_df = overlap_df[[vector_label, f'{str(self.x_axis_value)}_y', f'{str(self.y_axis_value)}_y']]
                vectors_2_df = vectors_2_df.rename(columns={
                    f'{str(self.x_axis_value)}_y': str(self.x_axis_value), 
                    f'{str(self.y_axis_value)}_y': str(self.y_axis_value), 
                    }
                )
                vectors_2_df = vectors_2_df.groupby([vector_label], sort=False).mean().reset_index()
                vectors_2_df['vector_field'] = vector_field_2
                if plot_nn: 
                    vectors_2_df['neighbours'] = overlap_df[['neighbours_y']]
                    vectors_2_df = vectors_2_df[:plot_nn]
                
                '''
                Plot overlap dfs
                '''
                for i, vectors_df in enumerate([ vectors_1_df, vectors_2_df ]):
                    text = vectors_df[vector_label] if (plot_labels) else None
                    if self.vector_label_char_length:
                        text = vectors_df[vector_label].apply( lambda x: x[: self.vector_label_char_length ] + "..." )
                    
                    labels = [vector_label, 'neighbours'] if plot_nn else [vector_label]
                    custom_data, hovertemplate = self._generate_hover_template(df=vectors_df, labels=labels)
                    scatter = go.Scatter(
                            name=str(vectors_df.vector_field.values[0]),
                            x=vectors_df[str(self.x_axis_value)],
                            y=vectors_df[str(self.y_axis_value)],
                            mode=plot_mode,
                            marker={"size": marker_size, "symbol": "circle"},
                            text=text,
                            textposition="middle center",
                            customdata=custom_data,
                            hovertemplate=hovertemplate,
                            marker_color=palette[i]
                        )
                    data.append(scatter)

                vectors_1_df = vectors_1_df.rename(columns={vector_label: vector_field_1})
                vectors_2_df = vectors_2_df.rename(columns={vector_label: vector_field_2})
                comparator_df = pd.concat([vectors_1_df, vectors_2_df], axis=1).drop(columns='vector_field')

            else:
                comparator_df = pd.concat([comparator_df, df], axis=1)


        """
        Plot title
        """
        # plot_title = plot_title.replace("</b>", f"\tEncoder: {self.encoder}</b>")
        if self.plot_unique:
            plot_title =  plot_title.replace("</b>", f"\tUnique: {self.plot_unique}</b>\t")
        if self.vector_label_char_length:
            plot_title = plot_title.replace("</b>", f"\tChar Length: {self.vector_label_char_length}<br></b>\t")
            
        """
        Generating figure and formatting title and legend
        """
        axes = {
            "title": plot_title,
            "showgrid": True,
            "zeroline": True,
            "showticklabels": False,
        }
        layout = go.Layout(scene={"xaxis": axes, "yaxis": axes})
        fig = go.Figure(data=data, layout=layout)
        fig.update_layout(
            title={
                "text": plot_title,
                "y": 0.9,
                "x": 0.05,
                "xanchor": "left",
                "yanchor": "top",
                "font": {"size": 12},
            },
            xaxis_title=str(self.x_axis_value),
            yaxis_title=str(self.y_axis_value),
        )
        fig.update_layout(
            legend={
                "font": {"size": 10},
                "itemwidth": 30,
                "tracegroupgap": 1,
            }
        )

        """
        Show line between vectors 1 and vectors 2
        """
        if overlap_only:
            vectors_1_pts = list(zip(vectors_1_df[str(self.x_axis_value)].values, vectors_1_df[str(self.y_axis_value)].values))
            vectors_2_pts = list(zip(vectors_2_df[str(self.x_axis_value)].values, vectors_2_df[str(self.y_axis_value)].values))
            
            for vectors_1_xy, vectors_2_xy in zip(vectors_1_pts, vectors_2_pts):
                fig.add_trace(
                    go.Scatter(
                        x=[ vectors_1_xy[0], vectors_2_xy[0] ],
                        y=[ vectors_1_xy[1], vectors_2_xy[1] ],
                        mode="lines",
                        line=go.scatter.Line(color="darkslategray", dash="dash", width=0.5),
                        showlegend=False,
                    )
                )  

        """
        Show trendline
        """
        projection_points = np.array( comparator_df[[str(self.x_axis_value), str(self.y_axis_value)]] )
        projection_points_min = projection_points.min()
        projection_points_max = projection_points.max()

        fig.add_trace(
            go.Scatter(
                x=[ projection_points_min, projection_points_max ],
                y=[ projection_points_min, projection_points_max ],
                mode="lines",
                line=go.scatter.Line(color="dimgray", width=0.6),
                showlegend=False
            )
        )
        self.fig = fig
        self.df = comparator_df

        return fig

        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-


import numpy as np
from typing import List, Dict, Tuple, Union


from .encode_utils import *
from constants import *


def vector_op(
    operation: VectorOperation, 
    vectors: List[np.ndarray],
    axis: int = 0
) -> np.ndarray:
    """
    Perform vector operations (op, v1, v2)
    """
    if operation == 'avg':
        return np.mean(vectors, axis=axis)
    elif operation == 'sum':
        return np.sum(vectors, axis=axis)
    elif operation == 'min':
        return np.min(vectors, axis=axis)
    elif operation == 'max':
        return np.max(vectors, axis=axis)
    elif operation == 'add':
        return np.add.reduce(vectors, axis=axis)
    elif operation == 'subtract':
        return np.subtract.reduce(vectors, axis=axis)
    elif not isinstance(operation, VectorOperation):
        raise ValueError(f'Not a valid vector operation {operation} {VectorOperation}')



def axes_to_vector(
    axes_labels: List[str], 
    vectors_unique_lut: dict,
) -> List[np.ndarray]:
    """
    Calculate the axes vectors defined by user
    """
    axes_vectors = []
    for label in axes_labels:
        f_vector = vectors_unique_lut.get(label)
        axes_vectors.append(f_vector)
    return axes_vectors




def formulae_to_vector(
    formulae: List[str], 
    dataset_info: dict,
    encoder: Union[None, str] = None
) -> List[np.ndarray]:
    """
    Calculate the vector of the math formulae defined by the user
    """
    formulae_vectors = []
    for formula in formulae:
        op = formula[0]
        v1 = formula[1]
        v2 = formula[2]
        if not op or op not in VectorOperation:
            raise ValueError(f'Invalid operator {VectorOperation}')
        # print(encoder)
        if encoder=='clip':
            v1_encoded = clip_encode_text(v1)
            v2_encoded = clip_encode_text(v2) 
            f_vector = vector_op(operation=op, vectors=[v1_encoded, v2_encoded])
        elif encoder=='context_avg':
            v1_encoded = get_search_word_vector(v1, dataset_info=dataset_info)
            v2_encoded = get_search_word_vector(v2, dataset_info=dataset_info)
            f_vector = vector_op(operation=op, vectors=[v1_encoded, v2_encoded])
        else:
            if dataset_info['vectors_lut'].get(v1) is None:
                raise ValueError(f"{v1} not in dataset {dataset_info['dataset_id']}")
            elif dataset_info['vectors_lut'].get(v1) is None:
                raise ValueError(f"{v2} not in dataset {dataset_info['dataset_id']}")
            
            f_vector = vector_op(operation=op, 
                                vectors=[ dataset_info['vectors_lut'].get(v1), 
                                        dataset_info['vectors_lut'].get(v2) ])
        
        formulae_vectors.append(f_vector)
        
    return formulae_vectors



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import pandas as pd

from abc import abstractmethod
from typing import List, Tuple, Dict, Any, Optional

from pathlib import Path
from dataclasses import dataclass
from typing import List
from collections import Counter

from doc_utils import DocUtils

from vectorops.datasets import data_dict

BASE_DIR = Path.cwd().resolve().parent
DATA_PATH = BASE_DIR / 'data'


## TODO: Refactor dataset class


@dataclass
class DatasetBase:
    vector_label: str
    vector_name: str
    dataset_id: str

    @abstractmethod
    def _load_data(self, docs: Dict[str, Any]):
        """Load data from docs"""
        raise NotImplementedError
    


@dataclass
class LocalDataset(DatasetBase, DocUtils):
    def __init__(self, dataset_info: dict, k: int = -1, norm: bool = True):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        self.dataset_info = dataset_info
        self.vector_label = dataset_info['vector_label']
        self.vector_name = dataset_info['vector_name']
        self.dataset_id = dataset_info['dataset_id']
        self.dataset_filename = dataset_info['dataset_filename']
        self.metadata_filename = dataset_info.get('metadata_filename')
        self.k = k
        self.norm = norm
        self._load_data()
        super().__init__(
            vector_label=self.vector_label, 
            vector_name=self.vector_name, 
            dataset_id=self.dataset_id
        )
    
    def _load_data(self):
        """
        Load vector dataset 
        """
        dataset_full = data_dict[self.dataset_filename][:self.k]

        if Path(DATA_PATH.joinpath(f'{self.metadata_filename}.csv')).exists():
            metadata = data_dict[self.metadata_filename]
        else:
            metadata_cols = [col for col in dataset_full.columns 
                            if not any(s in col for s in ['_vector_', '_id', 'insert_date_'])]
            metadata = dataset_full[metadata_cols]
            self.metadata_filename = self.dataset_filename.replace('max', 'metadata')
            metadata.to_csv(
                DATA_PATH.joinpath(f"{self.dataset_id}", f"{self.metadata_filename}.csv"), 
                encoding='utf-8', index=None
            )
        vectors = dataset_full[self.vector_name]
        vectors = np.array([x for x in vectors])
        self.vectors = vectors

        self.labels = metadata[[self.vector_label]]
        self.metadata = metadata
        if self.norm:
            self.vectors = self.vectors / np.linalg.norm(self.vectors, axis=1).reshape(-1, 1)

        labels_unique = pd.Series(metadata[self.vector_label].unique()).sort_values()
        vectors_unique = self._get_vector_unique(labels_unique)
        metadata_unique = metadata.loc[labels_unique.index]
        vectors_lut = {v: vectors_unique[i] for i, v in enumerate(labels_unique)}  
        self.labels_unique = labels_unique
        self.vectors_unique = vectors_unique
        self.metadata_unique = metadata_unique
        self.vectors_lut = vectors_lut


@dataclass
class DatasetManager(DatasetBase, DocUtils):
    def __init__(self, 
        k: int = -1, norm: bool = True, dataset_info: Optional[dict] = None
    ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        if dataset_info:
            self.dataset_info = dataset_info
            self.vector_label = dataset_info['vector_label']
            self.vector_name = dataset_info['vector_name']
            self.dataset_id = dataset_info['dataset_id']
            self.dataset_filename = dataset_info['dataset_filename']
            self.metadata_filename = dataset_info.get('metadata_filename')

        # self.docs = docs
        self.k = k
        self.norm = norm
        self._load_data()

        super().__init__(
                vector_label=self.vector_label, 
                vector_name=self.vector_name, 
                dataset_id=self.dataset_id
            )
        
    def _load_data(self):
        """
        Load vector dataset 
        """
        dataset_full = data_dict[self.dataset_filename]

        if Path(DATA_PATH.joinpath(f'{self.metadata_filename}.csv')).exists():
            metadata = data_dict[self.metadata_filename]
        else:
            metadata_cols = [col for col in dataset_full.columns 
                            if not any(s in col for s in ['_vector_', '_id', 'insert_date_'])]
            metadata = dataset_full[metadata_cols]
            self.metadata_filename = self.dataset_filename.replace('max', 'metadata')
            metadata.to_csv(
                DATA_PATH.joinpath(f"{self.dataset_id}", f"{self.metadata_filename}.csv"), 
                encoding='utf-8', index=None
            )

        vectors = dataset_full[self.vector_name]
        vectors = np.array([x for x in vectors])
        self.vectors = vectors

        self.labels = metadata[[self.vector_label]]
        self.metadata = metadata
        if self.norm:
            self.vectors = self.vectors / np.linalg.norm(self.vectors, axis=1).reshape(-1, 1)

        labels_unique = pd.Series(metadata[self.vector_label].unique()).sort_values()
        vectors_unique = self._get_vector_unique(labels_unique)
        metadata_unique = metadata.loc[labels_unique.index]
        vectors_lut = {v: vectors_unique[i] for i, v in enumerate(labels_unique)}  
        self.labels_unique = labels_unique
        self.vectors_unique = vectors_unique
        self.metadata_unique = metadata_unique
        self.vectors_lut = vectors_lut


    def __len__(self):
        return len(self.vectors)
    
    @property
    def len_unique(self):
        return len(self.vectors_unique)
    
    @property
    def shape(self):
        return self.vectors.shape
    
    @property
    def shape_unique(self):
        return self.vectors_unique.shape
    
    
    def _get_vector_unique(self, 
        labels_unique: pd.Series
    ) -> np.ndarray:
        '''
        Averaging unique vectors
        '''
        vectors_unique = self.vectors[labels_unique.index]
        labels = [l for llist in self.labels.values.tolist() for l in llist]
        c = Counter(labels)
        non_unique_labels = [k for k, v in c.items() if v>1]
        for i, v in enumerate(labels_unique):
            if v in non_unique_labels:
                vectors_unique[i] = self.get_search_word_vector(v)
        return vectors_unique


    def get_matching_items(self, 
        search_word: str, 
        metadata_cols: List = None,
        unique: bool = False
    ) -> pd.DataFrame:
        '''
        Get matching items of a given search word
        '''
        if unique:
            vector_label_lut = self.metadata_unique[self.vector_label].apply(lambda x : x.lower())
            search_word_mask = vector_label_lut.str.contains(search_word.lower())
            metadata_cols = [self.vector_label] + metadata_cols if metadata_cols else self.metadata_unique.columns
            matching_items = self.metadata_unique[metadata_cols].loc[search_word_mask]
        else:
            vector_label_lut = self.metadata[self.vector_label].apply(lambda x : x.lower())
            search_word_mask = vector_label_lut.str.contains(search_word.lower())
            metadata_cols = [self.vector_label] + metadata_cols if metadata_cols else self.metadata.columns
            matching_items = self.metadata[metadata_cols].loc[search_word_mask]
        return matching_items


    def get_search_word_vector(self,
        search_word: str,
        k: int = -1
    ) -> np.ndarray:
        '''
        Return average of search word vectors of a given search word
        '''
        vector_labels = self.labels[self.vector_label].apply(lambda x : x.lower())
        search_word_mask = vector_labels.str.contains(search_word.lower())
        search_word_vectors = self.vectors[search_word_mask][:k]
        return  np.mean(search_word_vectors, axis=0)




if __name__ == '__main__':
    dataset_info = {
        'vector_label': 'product_name',
        'vector_name': 'product_name_imagetext_vector_',
        'dataset_filename': 'ecommerce-6.uniq_id.7000.product_name_imagetext.7000.max',
        'metadata_filename': 'ecommerce-6.uniq_id.7000.product_name_imagetext.7000.metadata',
        'dataset_id':'ecommerce-6'
    }   

    dm = DatasetManager(dataset_info, data_dict)
    print(dm.vectors)
    print(len(dm))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Thursday, October 21st 2021, 11:45:36 pm
# Last Modified: Thursday, October 21st 2021,11:46:37 pm
#####

import numpy as np
import pandas as pd

from typing import List, Dict, Tuple, Union

from distance_utils import cosine_dis, euc_dis


def get_nn_idx(
    vectors: Union[pd.DataFrame, np.ndarray],
    selected_vec: Union[pd.Series, np.ndarray],
    distance_measure_mode: str = 'cosine'
) -> np.ndarray:
    '''
    Get nearest neighbours idx
    '''
    if (selected_vec.ndim > 1):
        selected_vec = selected_vec.iloc[0]

    def compare_vector(vector: np.ndarray):
        if distance_measure_mode == 'cosine':
            return cosine_dis(selected_vec, vector)
        elif distance_measure_mode == 'euclidean':
            return euc_dis(selected_vec, vector)

    if type(vectors) == pd.DataFrame:
        distance_map = vectors.apply(compare_vector, axis=1)
        return distance_map.sort_values().index
    elif type(vectors) == np.ndarray:
        distance_map = np.array([compare_vector(v) for v in vectors])
        return np.argsort(distance_map)
        # return np.argpartition(distances, range(0, k))[:k]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Friday, October 22nd 2021, 12:05:11 am
# Last Modified: Friday, October 22nd 2021,2:55:27 am
#####



from pathlib import Path
import pandas as pd

BASE_DIR = Path(__file__).resolve().parent.parent
PATH = Path(__file__).parent
DATA_PATH = BASE_DIR / 'data'

MAX_VECTORS = 7000

data_dict = {
    # 'ecommerce-6.uniq_id.7000.product_name_imagetext.7000.max': pd.read_csv(
    #     DATA_PATH.joinpath('ecommerce-6.uniq_id.7000.product_name_imagetext.7000.csv'), 
    #     encoding='utf-8'
    # ),
    # 'ecommerce-6.uniq_id.7000.product_name_imagetext.7000.metadata': pd.read_csv(
    #     DATA_PATH.joinpath('ecommerce-6.uniq_id.7000.product_name_imagetext.7000.metadata.csv'), 
    #     encoding='utf-8'
    # ),
    # 'ecommerce-6.product_name.5362.product_name_imagetext.5000': pd.read_csv(
    #     DATA_PATH.joinpath('ecommerce-6.product_name.5362.product_name_imagetext.5000.csv'), 
    #     encoding='utf-8'
    # ),
    # 'ecommerce-6.13058.max': pd.read_json(
    #     DATA_PATH.joinpath('ecommerce-6', 'ecommerce-6.13058.max.json'), 
    #     encoding='utf-8'
    # ),
    'ecommerce-6.7000.max': pd.read_json(
        DATA_PATH.joinpath('ecommerce-6', 'ecommerce-6.7000.max.json'), 
        encoding='utf-8'
    ),
    'ecommerce-6.7000.metadata': pd.read_csv(
        DATA_PATH.joinpath('ecommerce-6', 'ecommerce-6.7000.metadata.csv'), 
        encoding='utf-8'
    ),
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Friday, October 22nd 2021, 3:58:48 am
# Last Modified: Friday, October 22nd 2021,3:58:53 am
#####





#!/usr/bin/env python3
# -*- coding:utf-8 -*-


import numpy as np
import pandas as pd

import plotly.figure_factory as ff
import plotly.graph_objects as go

from dataclasses import dataclass

from typing import List, Optional

from vectorops.constants import *
from vectorops.utils.distance_utils import *
from vectorops.dataset_manager import DatasetManager
from vectorops.projection_manager import ProjectionManager


@dataclass 
class DistPlot(ProjectionManager):
    def __init__(self,  
            dm: DatasetManager,
            queries: List[str],
            items: List[str],
            encoder: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = None,
        ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        super().__init__(dm)

        self.queries = queries
        self.items = items
        self.encoder = encoder
        self.fig = go.Figure()

    
    def plot(
        self,
        encoder: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = None,
        trim: bool = False
    ):  
        """ 
        Plot Radar plot
        """
        if encoder: self.encoder = encoder
        hist_data = self._encode_items(items=self.items, encoder=self.encoder)
        
        if trim:
            hist_data_no_outliers = []
            for data in hist_data:
                data_mean, data_std = np.mean(data), np.std(data)
                cut_off = data_std * 3
                lower, upper = data_mean - cut_off, data_mean + cut_off
                hist_data_no_outliers.append([x for x in data if x > lower and x < upper])
            hist_data = hist_data_no_outliers
        
        self.fig = ff.create_distplot(hist_data, self.items, bin_size=.2)
        self.fig.update_layout(
            title_text=f"{self.dataset_id} <br>{self.vector_label}: {self.vector_name}  <br>Encoder: {self.encoder}",
            xaxis_title='Distribution',
            yaxis_title='Count',
        )
        self.fig.update_layout(legend_title_text=self.vector_label)

        return self.fig


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vectorops', 'vectorops.projection', 'vectorops.utils']

package_data = \
{'': ['*']}

install_requires = \
['RelevanceAI[notebook]>=0.17.0,<0.18.0',
 'document-utils>=1.3.0,<2.0.0',
 'numpy>=1.21.3,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'plotly>=5.3.1,<6.0.0',
 'scikit-learn>=1.0,<2.0',
 'scipy>=1.7.1,<2.0.0',
 'torch>=1.9.1,<2.0.0',
 'torchvision>=0.10.1,<0.11.0',
 'vectorhub[text-encoder-transformers]>=1.8.3,<2.0.0']

setup_kwargs = {
    'name': 'relevanceai-vectorops',
    'version': '0.1.1',
    'description': 'Vector Ops Exploration and Tooling',
    'long_description': None,
    'author': 'Charlene Leong',
    'author_email': 'charlene.leong@relevance.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-


import numpy as np
from scipy.spatial.distance import cdist

from typing import List, Dict, Tuple, Union


def cosine_dis(v1: Union[float, np.ndarray], v2: Union[float, np.ndarray]) -> np.ndarray:
    '''
    Calculates Cosine distance between two 1D arrays or two floats
    '''
    return 1 - (np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))


def cosine_sim(v1: Union[float, np.ndarray], v2: Union[float, np.ndarray]) -> np.ndarray:
    '''
    Calculates Cosine similarity between two 1D arrays or two floats
    '''
    return (np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))


def euc_dis(v1: Union[float, np.ndarray], v2: Union[float, np.ndarray]) -> np.ndarray:
    '''
    Calculates Euclidean distance between two 1D arrays or two floats
    '''
    return np.linalg.norm(v1-v2)


def cosine_sim_matrix(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    '''
    Calculates Cosine similarity matrix between each pair of the two collections of inputs
    '''
    return 1 - cdist(x, y, metric='cosine')


def cosine_dist_matrix(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    '''
    Calculates Cosine distance matrix between each pair of the two collections of inputs
    '''
    return cdist(x, y, metric='cosine')


# def min_max_norm(vectors: np.ndarray, min: float, max: float) -> np.ndarray:
#     '''
#     Performs min max normalisation on an array
#     '''
#     return np.frompyfunc(lambda x, min, max: (x - min) / (max - min), 3, 1)


min_max_norm = np.frompyfunc(lambda x, min, max: (x - min) / (max - min), 3, 1)


#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
import pandas as pd

import plotly.express as px
import plotly.graph_objects as go

from dataclasses import dataclass

from typing import List, Optional

from vectorops.constants import *
from vectorops.utils.distance_utils import *
from vectorops.dataset_manager import DatasetManager
from vectorops.projection_manager import ProjectionManager


@dataclass 
class RadarProjection(ProjectionManager):
    def __init__(self,  
            dm: DatasetManager,
            queries: List[str],
            items: List[str],
            encoder: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = None,
        ):    
        """
        Initialise dataset manager from dataset info and data dict
        """
        super().__init__(dm)

        self.queries = queries
        self.items = items
        self.encoder = encoder
    
        self.fig = go.Figure()


    def plot(self,
        encoder: Optional[Union[Base2Vec, SentenceTransformer2Vec]] = None,
        trim: bool = True
    ):  
        """
        Plot Radar plot
        """
        if encoder: self.encoder = encoder
        theta_vectors = self._encode_items(items=self.items, encoder=self.encoder)

        plot_title = f"{self.dataset_id} <br>{self.vector_label}: {self.vector_name}  <br>Encoder: {self.encoder}"
        self.fig.update_layout(title=plot_title)
        
        r_min = 1
        r_max = 0
        for query in self.queries:
            qv = self._clip_encode_text(query, norm=self.norm)
            r = [cosine_sim(tv, qv) for tv in theta_vectors]
            r_min = min(r_min, min(r))
            r_max = max(r_max, max(r))
            self.fig.add_trace(go.Scatterpolar(
                r= r,
                theta=self.items,
                fill='toself',
                name=query,
                connectgaps=True,
                )
            )
        if trim:
            self.fig.update_layout(
                polar=dict(
                    radialaxis=dict(
                    visible=True,
                    range=[r_min, r_max]
                    )),
                )
            
        return self.fig


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Thursday, October 21st 2021, 11:24:13 pm
# Last Modified: Friday, October 22nd 2021,4:22:22 am
#####

import numpy as np
import pandas as pd

import plotly.express as px
import plotly.graph_objects as go

from typing import List, Dict, Tuple, Union

from nn_utils import get_nn_idx

from constants import *


## TODO: Refactor into class

def plot_projection_df(
    projection_matrix: np.ndarray,
    axes_matrix: np.ndarray,
    x_axis_value: VectorOperation,
    y_axis_value: VectorOperation,
    dataset_info: dict,
    plot_nn: Union[None, int] = None, 
    plot_labels: bool = False
):
    """
    Takes projection matrix and returns df
    """
    
    def build_projection_df():
        """
        Build project df from projection matrix
        """
        if not plot_nn:
            labels_df = pd.DataFrame(dataset_info['vectors_lut'].keys(), 
                                        columns=[dataset_info['vector_label']])
            data_df = pd.DataFrame(projection_matrix, 
                                   columns=[str(x_axis_value), str(y_axis_value)])
            legend = None
        else:
            search_vector_nn_idx_1 = get_nn_idx(vectors=dataset_info['vectors'], 
                                                selected_vec=axes_matrix[0])
            nn_1_labels = dataset_info['metadata'][[dataset_info['vector_label']]].loc[search_vector_nn_idx_1][:plot_nn]
            nn_1_labels[''] = 'x'
            
            search_vector_nn_idx_2 = get_nn_idx(vectors=dataset_info['vectors'], 
                                                selected_vec=axes_matrix[1])
            nn_2_labels = dataset_info['metadata'][[dataset_info['vector_label']]].loc[search_vector_nn_idx_2][:plot_nn]
            nn_2_labels['neighbours'] = 'y'
            
            labels_df =  pd.concat([nn_1_labels, nn_2_labels], axis=0).reset_index()
            data_df = pd.DataFrame(projection_matrix[labels_df.index], 
                                    columns=[str(x_axis_value), str(y_axis_value)])
            legend = 'neighbours'
            
        return pd.concat([labels_df, data_df], axis=1), legend
    
    

    def clean_vector_op_label(axis_value: VectorOperation) -> str:
        """
        Converting operation tuple to axis label for px scatter
        """
        return ''.join(c for c in str(axis_value).replace(',', ',<br>') if c not in set('",()\]\''))
    
    
    df, legend = build_projection_df()
    text = df[dataset_info['vector_label']] if (plot_labels) else None

    labels = {}
    for axis in [ str(x_axis_value), str(y_axis_value) ]:
        if axis.op is None:
            labels[str(axis)] = axis.query
        else:
            labels[str(axis)] = '<br>'.join(axis.to_list())

    fig = px.scatter(df, x=str(x_axis_value), y=str(y_axis_value), color=legend, text=text,
                        title=f"{dataset_info['dataset_id']}: {len(df)} points<br>{dataset_info['vector_label']}: {dataset_info['vector_name']}",
                        labels=labels
                    )
    
    """
    Adding product name to hover data
    """
    fig.update_traces(customdata=df[dataset_info['vector_label']])
    fig.update_traces(hovertemplate='%{customdata}')
    fig.update_traces(
        hovertemplate="<br>".join([
            "X: %{x}",
            "Y: %{y}",
            "Label: %{customdata}",
        ])
    )
    fig.update_traces(
        hovertemplate="<br>".join([
            "X: %{x}",
            "Y: %{y}",
            "Label: %{customdata}",
        ])
    )
    fig.update_layout(
        title_font_size=10,
    )   


    """
    Show trendline
    """
    projection_matrix = np.array(df[[str(x_axis_value), str(y_axis_value)]])
    fig.add_trace(
        go.Scatter(
            x=[projection_matrix.min(), projection_matrix.max()],
            y=[projection_matrix.min(), projection_matrix.max()],
            mode="lines",
            line=go.scatter.Line(color="gray"),
            showlegend=False)
    )

#     fig.show()
    return fig, df


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Friday, October 22nd 2021, 12:41:08 am
# Last Modified: Friday, October 22nd 2021,3:59:50 am
#####


import numpy as np
from dataclasses import dataclass
from typing import List, Dict, Literal, Tuple, Union

from vectorhub.base import Base2Vec
from vectorhub.encoders.text.sentence_transformers import SentenceTransformer2Vec
from vectorhub.bi_encoders.text_image.torch import ClipText2Vec

VectorOperation = Literal['avg', 'sum', 'min', 'max', 'add', 'subtract', None]


@dataclass
class VectorFormula:
    op: VectorOperation
    queries: List[str]

    def __len__(self):
        return len(self.queries)

    def __repr__(self):
        return f'{self.op}({self.queries})'

    def __str__(self):
        return f'{self.op} ( {", ".join(self.queries)} )'
        

    @property
    def query(self):
        if self.op is None:
            return self.queries[0] 

    def to_list(self):
        return [self.op] + self.queries
        


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#####
# Author: Charlene Leong charleneleong84@gmail.com
# Created Date: Friday, October 22nd 2021, 1:12:15 am
# Last Modified: Friday, October 22nd 2021,4:21:19 am
#####

import numpy as np
import clip
import torch


device = "cuda" if torch.cuda.is_available() else "cpu"
model, preprocess = clip.load("ViT-B/32", device=device)


def clip_encode_text(sequence: str, norm: bool = False) -> np.ndarray:
    '''Computes the text vectors for text sequence'''
    with torch.no_grad():
        text_encoded = model.encode_text(clip.tokenize(sequence).to(device))
        if norm: text_encoded /= text_encoded.norm(dim=-1, keepdim=True)
    return text_encoded.cpu().numpy()[0]



def get_search_word_vector(
    search_word: str,
    dataset_info: dict
) -> np.ndarray:
    '''
    Return average of search word vectors of a given search word
    '''
    vector_labels = dataset_info['labels'][dataset_info['vector_label']].apply(lambda x : x.lower())
    search_word_mask = vector_labels.str.contains(search_word.lower())
    search_word_vectors = dataset_info['vectors'][search_word_mask]
    if search_word_vectors.shape[0] >= 5:
        search_word_vectors = search_word_vectors[:5]
    return  np.mean(search_word_vectors, axis=0)